
import os
from positron_common.job_api.funding_envs_images import *
from positron_common.cli.console import console, ROBBIE_BLUE, ROBBIE_DEFAULT
from prompt_toolkit import prompt
from prompt_toolkit.styles import Style
from prompt_toolkit.validation import Validator
from prompt_toolkit.completion import WordCompleter, PathCompleter
from positron_common.config import PositronJob
from positron_common.config import PositronJobConfig
from positron_common.enums import JobRunType
from positron_common.image_name import get_auto_image_name_and_cluster


# Prompts the user for FGs, environmetns, etc.
# Queries the backend for the FundingSources and Environments.
def prompt_and_build_positron_job_config(preserve: bool = False
) -> PositronJobConfig:

    try:
        pj = PositronJob()

        style = Style.from_dict({
            'completion-menu.completion': 'bg:#008888 #ffffff',
            'completion-menu.completion.current': 'bg:#00aaaa #000000',
        })
        # user can enter a name

        message = [
            (ROBBIE_BLUE, "Please enter a custom run name ["),
            (ROBBIE_DEFAULT, "let Robbie choose"),
            (ROBBIE_BLUE, "]:" ),
        ]
        name_choice = prompt(message=message, style=style)
        if len(name_choice):
            pj.name = name_choice

        # fetch the names and ids
        fs = FundingSources.load()

        # validate the user input
        def is_fs_valid(text):
            return text == "" or text in fs.menu_items()
        fs_validator = Validator.from_callable(is_fs_valid, error_message='Please select a valid funding source.')

        message = [ 
            (ROBBIE_BLUE, "Select how to bill your job ["),
            (ROBBIE_DEFAULT, "Personal tokens"),
            (ROBBIE_BLUE, "]:" ),
        ]
        fs_choice = prompt(message=message, completer=WordCompleter(fs.menu_items()), style=style, validator=fs_validator)

        if len(fs_choice):
            fs_id = fs.id_from_menu_item(fs_choice)
            pj.funding_selection = "User Selected"
        else:
            fs_id = fs.default_funding_source_id()
            pj.funding_selection = "User Selected - Default"
            
        pj.funding_group_id = fs_id

        #
        # Envrionments
        #
        # are there any environments in this funding source?
        envs = Environments.load(pj.funding_group_id)

        # validate the user input
        def is_env_valid(text):
            return text == "" or text in envs.menu_items()
        env_validator = Validator.from_callable(is_env_valid, error_message='Please enter a valid environment.')

        if envs:
            message = [ 
                (ROBBIE_BLUE, "Select your preferred hardware ["),
                (ROBBIE_DEFAULT, "None" if not fs.default_env_name_from_fs_id(fs_id) else fs.default_env_name_from_fs_id(fs_id)), # this is null for personal
                (ROBBIE_BLUE, "]:" ),
            ]
            env_choice = prompt(message=message, completer=WordCompleter(envs.menu_items()), style=style, validator=env_validator)
        else:
            # no environments for the user oh well
            console.print(f"[red]Error your funding sources: {fs_choice} has no approved hardware, please contact 'support@robbie.run'")
            return None

        if len(env_choice):
            pj.environment_id = envs.id_from_menu_item(env_choice)
            pj.environment_selection = "User Selected"
        else:
            # choose the default, if available
            if fs.default_env_id_from_fs_id(fs_id) == None:
                console.print(f"[red] Error funding source: {fs_choice} has no default hardware and you didn't specify any.")
                return None
            else:
                pj.environment_id = fs.default_env_id_from_fs_id(fs_id)
                pj.environment_selection = "User Selected - Default"

        #
        # Images
        #
        images = Images.load(pj.funding_group_id, pj.environment_id)

        def is_image_valid(text):
            return text == "" or text in images.menu_items()
        
        image_validator = Validator.from_callable(is_image_valid, error_message='Please choose a valid image.')

        if images:
            message = [ 
                    (ROBBIE_BLUE, "Select your preferred image["),
                    (ROBBIE_DEFAULT, "auto-select"),
                    (ROBBIE_BLUE, "]:" ),
                ]
            image_choice = prompt(message=message, completer=WordCompleter(images.menu_items()), style=style, validator=image_validator)

            if len(image_choice):
                # the user hit tab and selected an image
                pj.image = images.name_from_menu_item(image_choice)
                # we just get the cluster type from the environment
                pj.cluster = envs.cluster_type_from_env_id(pj.environment_id)
                pj.image_selection = "User Selected"
            else:
                # the user just hit <return>, so auto-select
                pj.image, pj.cluster = get_auto_image_name_and_cluster(pj.funding_group_id, pj.environment_id)
                console.print('Auto-selecting image:', style=ROBBIE_BLUE, end="")
                console.print(f'{pj.image} on cluster {pj.cluster}')
                pj.image_selection = "Automatically Selected"
        else:
            # no environments for the user oh well
            console.print(f"[red]Failed to get images, please contact 'support@robbie.run'")
            return None
        #
        # Dependencies
        #
        def is_dependency_valid(text):
            return text == "" or text in ["auto-capture", "./requirements.txt"]
        
        dependency_validator = Validator.from_callable(is_dependency_valid, error_message='Please choose a valid selection.')
        message = [ 
                (ROBBIE_BLUE, "Dependencies["),
                (ROBBIE_DEFAULT, "none"),
                (ROBBIE_BLUE, "]:" ),
            ]
        deps = prompt(message=message, completer=WordCompleter(["auto-capture", "./requirements.txt" ]), style=style, validator=dependency_validator)
        if len(deps):
            pj.dependencies = deps


        # 
        # Max tokens
        #
        
        def is_max_token_valid(text):
            return text == "" or (text.isdigit() and int(text) >=1 and int(text) < 10000)
        
        max_token_validator = Validator.from_callable(is_max_token_valid, error_message='Please enter a number between 1 and 10000.')
        

        message = [ 
                (ROBBIE_BLUE, "Maximum tokens to consume ["),
                (ROBBIE_DEFAULT, "none"),
                (ROBBIE_BLUE, "]:" ),
            ]
        text = prompt(message=message, style=style, validator=max_token_validator)
        if len(text):
            pj.max_tokens = int(text)
        

        # 
        # Max duration
        #
        def is_max_duration_valid(text):
            return text == "" or (text.isdigit() and int(text) >=1 and int(text) < 1000)
        duration_validator = Validator.from_callable(is_max_duration_valid, error_message='Please enter a valid duration between 1 and 1000.')


        message = [ 
                (ROBBIE_BLUE, "Max duration in minutes ["),
                (ROBBIE_DEFAULT, "none"),
                (ROBBIE_BLUE, "]:" ),
            ]
        text = prompt(message=message,style=style, validator=duration_validator)
        if len(text):
            pj.max_time = int(text)

        message = [ 
                (ROBBIE_BLUE, "Specify the directory contents to send to the remote machine ["),
                (ROBBIE_DEFAULT, os.getcwd()),
                (ROBBIE_BLUE, "]:" ),
            ]
        text = prompt(message=message, style=style)
        if len(text):
            pj.workspace_dir = text
        else:
            pj.workspace_dir = os.getcwd()

        # 
        # Environment variables
        #
        if not preserve:
            first_pass = True
            while True:
                var_name = prompt('Environment variable name (Enter a <blank> line to go to the next step):',style=Style.from_dict({'prompt': ROBBIE_BLUE}))
                if not var_name:
                    break
                var_value = prompt(f'Value for {var_name} (hint= Enter a <blank> line to use local machine value):', style=style)
                if first_pass:
                    pj.env = {}
                    first_pass = False
                pj.env[var_name] = var_value
        # 
        # Commands - loop through and capture the commands
        #
        if not preserve:
            first_pass = True
            while True:
                cmd = prompt('Enter command to run (Enter a <blank> line when you are done entering commands):', style=Style.from_dict({'prompt': ROBBIE_BLUE}))
                if not cmd:
                    break
                if first_pass:
                    pj.commands = []
                    first_pass = False
                pj.commands.append(cmd)

        pj.job_type = JobRunType.BASH_COMMAND_RUNNER

        return PositronJobConfig(version="1.0", python_job=pj)
    
    except Exception as e:
       print(f"Error: {e}")
       return None


# Class and singleton builds a list of tuples from the DB results
class FundingSources: 
    is_init: bool = False
    my_fs: dict

    def __init__(self, fs_arg: dict):
        if self.is_init:
            raise ValueError('FundingSources.load() already initialized')
        else:
            self.init = True
            self.my_fs= fs_arg

    @staticmethod
    def load():
        fs = list_funding_sources()
        if len(fs) == 0:
            return None
        # Loop through and add a custom "menu" item to each dict 
        for key, val in fs.items(): 
                val[FS_MENU] = f'{val[FS_NAME]} ({val[FS_TOKENS]} tokens available)'
        return FundingSources(fs)
        
    # Prompt toolkit needs a list of strings to display in the menu 
    def menu_items(self) -> list: 
        ret_list: list = []
        for key, val in self.my_fs.items():
            # just show names
            ret_list.append(val[FS_MENU])
        return ret_list

    # Return 'funding_group_id' using the val returned from session.prompt() 
    def id_from_menu_item(self, menu_item: str) -> str:
        for key, val in self.my_fs.items():
            if (val[FS_MENU] == menu_item):
                return val.get(FS_ID)
        return None

    def default_env_id_from_menu_item(self, menu_item: str) -> str:
        for key, val in self.my_fs.items():
            if (val[FS_MENU] == menu_item):
                return val.get(FS_DEF_ENV_ID)
        return None
    
    def default_env_name_from_menu_item(self, menu_item: str) -> str:
        for key, val in self.my_fs.items():
            if (val[FS_MENU] == menu_item):
                return val.get(FS_DEF_ENV_NAME)
        return None
    
    def default_env_name_from_fs_id(self, id: str) -> str:
        for key, val in self.my_fs.items():
                if (val[FS_ID] == id):
                    return val.get(FS_DEF_ENV_NAME)
        return None
    
    def default_env_id_from_fs_id(self, id: str) -> str:
        for key, val in self.my_fs.items():
                if (val[FS_ID] == id):
                    return val.get(FS_DEF_ENV_ID)
        return None
    
    def default_env_id(self) -> str:
        for key, val in self.my_fs.items():
            if (val[FS_TYPE] == FS_PERSONAL_TYPE):
                return val.get(FS_DEF_ENV_ID)
        return None
    
    def default_funding_source_id(self) -> str: 
        for key, val in self.my_fs.items():
            if (val[FS_TYPE] == FS_PERSONAL_TYPE):
                return val.get(FS_ID)
        return None
    
    def default_image_name(self, id: str) -> str:
        for key, val in self.my_fs.items():
            if (val[FS_ID] == id):
                return val.get(FS_DEFAULT_IMAGE_NAME)
        return None
    
    def default_image_id(self, id: str) -> str:
        for key, val in self.my_fs.items():
            if (val[FS_ID] == id):
                return val.get(FS_DEFAULT_IMAGE_ID)
        return None
    


# singleton for Environments
class Environments: 
    is_init: bool = False
    my_envs: dict

    def __init__(self, env_arg):
         if self.is_init:
            raise ValueError('Environments.load() already initialized')
         else:
            self.my_envs = env_arg
            self.is_init = True

    @staticmethod
    def load(fs_id: str):
        envs = list_environments(fs_id)
        if len(envs) == 0:
            return None
        for key, val in envs.items():
            val[ENV_MENU_ITEM] = f"{val['environmentName']} ({val['tokensPerHour']} Tokens/Hour)" # shows in menu
        return Environments(envs)

    def menu_items(self) -> list: 
        menu_list = []
        for key, val in self.my_envs.items():
            menu_list.append(val[ENV_MENU_ITEM])
        return menu_list

    def id_from_menu_item(self, menu_item: str) -> str:
        for key, val in self.my_envs.items():
            if (val[ENV_MENU_ITEM] == menu_item):
                return val.get(ENV_ID)
        return None

    def tokens_per_hour(self, env_id: str) -> str:
        for key, val in self.my_envs.items():
            if (val[ENV_ID] == env_id):
                return val.get(ENV_TPH)
        return None
    
    def cluster_type_from_env_id(self, env_id: str) -> str:
        for key, val in self.my_envs.items():
            if (val[ENV_ID] == env_id):
                return val.get(ENV_CLUSTER_TYPE)
        return None
        


# singleton for Environments
class Images: 
    is_init: bool = False
    my_images: dict

    def __init__(self, image_arg):
         if self.is_init:
            raise ValueError('Images.load() already initialized')
         else:
            self.my_images = image_arg
            self.is_init = True

    @staticmethod
    def load(fs_id: str, env_id: str):
        images = list_images(fs_id, env_id)
        if len(images) == 0:
            return None
        for key, val in images.items():
            val[IMAGE_MENU_ITEM] = f"{val[IMAGE_NAME]}" # shows in menu
        return Images(images)

    def menu_items(self) -> list: 
        menu_list = []
        for key, val in self.my_images.items():
            if val.get(IMAGE_DELETED) == False:
                menu_list.append(val[IMAGE_MENU_ITEM])
        return menu_list

    def name_from_menu_item(self, menu_item: str) -> str:
        for key, val in self.my_images.items():
            if (val[IMAGE_MENU_ITEM] == menu_item):
                return val.get(IMAGE_NAME)
        return None
        



from dataclasses import dataclass, asdict
from positron_common.exceptions import DeserializationException
from typing import Union
import json
import sys

# this function is used exclusively in the MetaData class
def get_python_version():
    return f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"

@dataclass
class FunctionHashes:
    func_hash: Union[None, str] = None
    args_hash: Union[None, str] = None
    kwargs_hash: Union[None, str] = None
    result_hash: Union[None, str] = None
    exception_hash: Union[None, str] = None

# TODO: should we convert to a yaml format and stringify/parse the same as JobConfig?
@dataclass
class MetaData:
    """Metadata about the serialized data or functions."""

    sha256_hash: FunctionHashes
    version: str = "2023-04-24"
    python_version: str = get_python_version()
    serialization_module: str = "cloudpickle"

    def to_json(self):
        return json.dumps(asdict(self)).encode()

    @staticmethod
    def from_json(s):
        try:
            obj = json.loads(s)
        except json.decoder.JSONDecodeError:
            raise DeserializationException("Corrupt metadata file. It is not a valid json file.")

        sha256_hash = obj.get("sha256_hash")
        metadata = MetaData(sha256_hash=sha256_hash)
        metadata.version = obj.get("version")
        metadata.python_version = obj.get("python_version")
        metadata.serialization_module = obj.get("serialization_module")

        if not sha256_hash:
            raise DeserializationException(
                "Corrupt metadata file. SHA256 hash for the serialized data does not exist. "
                "Please make sure to install SageMaker SDK version >= 2.156.0 on the client side "
                "and try again."
            )

        if not (
            metadata.version == "2023-04-24" and metadata.serialization_module == "cloudpickle"
        ):
            raise DeserializationException(
                f"Corrupt metadata file. Serialization approach {s} is not supported."
            )

        return metadata


import importlib.metadata
import sys
import os
from positron_common.cli.logging_config import logger

def get_version():
    version = importlib.metadata.version('robbie')
    return version

def _current_python_version():
    """Returns the current python version where program is running"""

    return f"{sys.version_info.major}.{sys.version_info.minor}".strip()

def get_default_workspace_dir():
    cwd = os.getcwd()
    return cwd

def is_notebook():
    """ Returns True if running in a notebook, False otherwise """
    try:
        from IPython import get_ipython
        ip = get_ipython()
        if '__vsc_ipynb_file__' in ip.user_ns:
            logger.debug("Visual Studio Code detected!") 
            return True
        elif 'JPY_SESSION_NAME' in os.environ:
            logger.debug("Generic Jupyter Notebook detected!")
            return True
        elif 'google.colab' in str(ip):
            logger.debug("Google Colab detected!")
            return True
        else:
            logger.debug("No notebook detected")
            return False
    except AttributeError:
        logger.debug("ImportError, returning false")
        return False
    except ImportError:
        logger.debug("ImportError, returning false")
        return False
    except NameError:
        logger.debug("get_ipython() failed, not running in notebook")
        return False
    logger.debug("not running in notebook")
    return False
    
# singleton to check if we are running in a notebook
_nb = is_notebook()

#sys.exit() codes - only return if not running in a jupyter notebook
SUCCESS=0
FAILURE=1

def _exit_by_mode(val):
    """
    This function helps achieve two things:
    - If the code is not running in a notebook, it returns the value via sys.exit(). Users can write scripts with "robbie run" and check return values
    - If the code is running in a notebook, it returns None, since sys.exit() will kill the kernel in a notebook
    """
    # logger.debug(f'Exiting..._nb={_nb}, value={val}')
    logger.debug(f'Exiting...')
    sys.exit(val) if not _nb else None


# Sentinel object for undefined
undefined = object()



import platform
from rich.console import Console
from rich import box
from rich.text import Text
from rich.panel import Panel

# Website colors
ROBBIE_BLUE='#41a7ff'
ROBBIE_DEFAULT='#ced7bd'
ROBBIE_ORANGE='#ff3a24'
ROBBIE_GREY='#f2f2f2'
ROBBIE_YELLOW='#f5f543'

# Used to print nice message to the end user, this is not a logger. See `logging_config.py` for logging.
console = Console()

def print_boxed_messages(message: str, title: str):
    """
    Prints a message in a box with a title.
    """
    text = Text()
    text.append(message)
    console.print(Panel(
        text,
        box=box.ROUNDED,
        # padding=(1, 2),
        title = Text(title, style=ROBBIE_BLUE),
        border_style=ROBBIE_ORANGE,
    ))



from enum import Enum
from importlib.resources import files
from positron_common.cli.logging_config import logger
from positron_common.cli.console import console

class EnvType(Enum):
    LOCAL = "local"
    DEV = "dev"
    ALPHA = "alpha"
    BETA = "beta"

class BuildEnvFile():
  """
  Plain text file that reflects which env this package was build for.
  This will drive the default values that can be overridden via config options.
  """
  value: EnvType = EnvType.DEV

  def __init__(self):
    self.value = self.get_type()

  def get_text(self) -> str:
    build_env_resource = files('positron_common').joinpath('build_env')
    if (not build_env_resource.is_file()):
      logger.debug(f'"build_env" file not found, defaulting to "{self.value.value}"')
      return self.value.value
    try:
      contents = build_env_resource.read_text().strip()
    except Exception as e:
      logger.debug(f'Error reading build_env file: {e}, defaulting to "{self.value.value}"')
      return self.value.value
    if not contents:
      logger.debug(f'Empty build_env file, defaulting to "{self.value.value}"')
      return self.value.value
    return contents

  def get_type(self) -> EnvType:
    try:
      return EnvType(self.get_text())
    except ValueError as e:
      console.print(f"[yellow]There seems to be a problem with the build: {e}[/yellow]")
      console.print(f"[yellow]Defaulting to {self.value.value}[/yellow]")
      return self.value

build_env: EnvType = BuildEnvFile().value
"""
The type of env this package was built for.
Used when selecting default values for the package.
"""



import boto3
from botocore.client import BaseClient
from .runner_env import runner_env

def create_client() -> BaseClient:
    robbie_access_key_id = runner_env.ROBBIE_RUNNER_AWS_ACCESS_KEY_ID
    robbie_secret_access_key = runner_env.ROBBIE_RUNNER_AWS_SECRET_ACCESS_KEY

    # Initialize CloudWatch client
    if robbie_access_key_id and robbie_secret_access_key:
      # If the ROBBIE_RUNNER-prefixed credentials exist, use them
        session = boto3.Session(
            aws_access_key_id=robbie_access_key_id,
            aws_secret_access_key=robbie_secret_access_key
        )
        return session.client('logs', region_name=runner_env.REGION)

    return boto3.client('logs', region_name=runner_env.REGION)



from pydantic import BaseModel
from typing import List, Optional
from positron_common.cli.console import console, print_boxed_messages
from rich.text import Text

class PositronCLIArgs(BaseModel):
    """
    Positron CLI command line arguments.
    """
    is_init: bool = False
    name: Optional[str] = None
    local: bool = False
    deploy: bool = False
    stream_stdout: bool = False
    job_args: Optional[List[str]] = None
    skip_prompts: bool = False
    monitor_status: bool = False
    commands_to_run: Optional[str] = None
    interactive: bool = False
    create_only: bool = False
    results_from_job_id: str = ""
    download: Optional[str] = None
    local_path: Optional[str] = None
    auto_capture_deps: bool = False


    def init(self,
        name: Optional[str] = None,
        local: bool = False,
        deploy: bool = False,
        stream_stdout: bool = False,
        job_args: Optional[List[str]] = None,
        skip_prompts: bool = False,
        monitor_status: bool = False,
        commands_to_run: Optional[str] = None,
        interactive: bool = False,
        create_only: bool = False,
        results_from_job_id: str = "",
        download: Optional[str] = None,
        local_path: Optional[str] = None,
        auto_capture_deps: bool = False
    ):
        if self.is_init:
            raise ValueError('CLI Args already initialized')
        
        self.name = name
        self.local = local
        self.deploy = deploy
        self.stream_stdout = stream_stdout
        self.job_args = job_args
        self.is_init = True
        self.skip_prompts=skip_prompts
        self.monitor_status=monitor_status
        self.commands_to_run = commands_to_run
        self.interactive = interactive
        self.create_only = create_only
        self.results_from_job_id = results_from_job_id
        self.download = download
        self.local_path = local_path
        self.auto_capture_deps = auto_capture_deps

    def to_string(self, include_title: bool = False) -> str:
        message = f"""- name: {self.name}
- local: {self.local}
- deploy: {self.deploy}
- stream_stdout: {self.stream_stdout}
- job_args: {self.job_args}
- skip_prompts: {self.skip_prompts}
- monitor_status: {self.monitor_status}
- commands_to_run: {self.commands_to_run}
- interactive: {self.interactive}
- create_only: {self.create_only}
- results_from_job_id: {self.results_from_job_id}
- download: {self.download}
- local_path: {self.local_path}
- auto_capture_deps: {self.auto_capture_deps}
- is_init: {self.is_init}"""
        
        if include_title:
            return f"========== CLI Arguments (args) ==========\n{message}"
        else:
            return message

    


#
# Export global (singleton)
#
args = PositronCLIArgs()
"""
Global CLI arguments singleton, make sure you call init() before using it.
"""



import yaml
from typing import Optional
from pydantic import BaseModel
import os
from positron_common.constants import APP_HOME_DIR
from positron_common.cli.logging_config import logger
from positron_common.cli.console import console, print_boxed_messages

DEFAULT_USER_CONFIG_PATH = f"{APP_HOME_DIR}/config.yaml"
REMOTE_FUNCTION_SECRET_KEY_NAME = 'REMOTE_FUNCTION_SECRET_KEY'

class UserConfigFile(BaseModel):
    """
    The user configuration file for the Robbie CLI.
    """
    user_auth_token: Optional[str] = None
    backend_api_base_url: Optional[str] = None
    backend_ws_base_url: Optional[str] = None
    
    def load_config(self):
        """
        Loads the user configuration from the user's home directory.
        """
        if not os.path.exists(DEFAULT_USER_CONFIG_PATH):
            logger.debug(f'user config file not found: `{DEFAULT_USER_CONFIG_PATH}`')
            return

        try:
            with open(DEFAULT_USER_CONFIG_PATH, 'r') as user_config_file:
                user_config_dict = yaml.safe_load(user_config_file)
                self.__dict__.update(**user_config_dict)
        except Exception as e:
            console.print(f'[red]Error loading user configuration! {str(e)}[/red]')
            console.print(f'[yellow]Ignoring user config file: `{DEFAULT_USER_CONFIG_PATH}`[/yellow]')
    
    def write(self):
        """
        Write the user configuration to the user's home directory.
        """
        data = self.model_dump(exclude_none=True)
        os.makedirs(os.path.dirname(DEFAULT_USER_CONFIG_PATH), exist_ok=True)
        with open(DEFAULT_USER_CONFIG_PATH, 'w') as user_config_file:
            yaml.dump(data, user_config_file)
    
    def to_string(self, include_title: bool = False) -> str:
        """
        Print the user configuration to the console.
        """
        message = f"""- User Auth Token: {self.user_auth_token}
- Backend API Base URL: {self.backend_api_base_url}
- Backend WS Base URL: {self.backend_ws_base_url}"""
        if include_title:
            return f"========== User Config (user_config) ==========\n{message}"
        else:
            return message
 

user_config = UserConfigFile()
"""
The user configuration file for the Robbie CLI.
"""

if os.getenv('POSITRON_CLOUD_ENVIRONMENT', False):
    logger.debug("Cloud environment detected, skipping config file load.")
else:
    logger.debug('loading user config')
    user_config.load_config()



import typer
from typing import Optional
from logging import _nameToLevel
from positron_common.cli.logging_config import set_log_level

def log_level(
    loglevel: Optional[str] = typer.Option(
        None, help="Set the logging level [CRITICAL,FATAL,ERROR, WARNING, INFO, DEBUG, NOTSET]"
    ),
):
    set_log_level(loglevel)


from rich import print
from positron_common.cli.interactive import prompt_and_build_positron_job_config
from positron_common.constants import JOB_CONF_YAML_PATH
from positron_common.observability.main import track_command_usage

@track_command_usage("configure")
def configure() -> None:
    """
    Build a Robbie job configure file (job_config.yaml) interactively.

    """
    print("Please follow the prompts to build a 'job_config.yaml' in your current directory.")
    print("Values in brackets [] are default, use the <tab> key to see a menu of options.\n")
    
    cfg = prompt_and_build_positron_job_config()

    if cfg:
        cfg.write_to_file(filename=JOB_CONF_YAML_PATH)
        print(f"[green] Successfully wrote config file {JOB_CONF_YAML_PATH}'")
        print(f"[yellow]---------- Contents ----------")
        print(cfg)
        print(f"[yellow]---------- End ----------")




    



import signal
import json
import time
from rich.prompt import Confirm
from positron_common.aws.s3_presigned_handler import S3PresignedHandler
from positron_common.deployment.stored_function import StoredFunction
from positron_common.utils import get_default_workspace_dir, _exit_by_mode, SUCCESS, FAILURE, _nb
from positron_common.exceptions import RemoteFunctionException, RemoteCallException
from positron_common.config import PositronJob
from positron_common.enums import JobStatus, failed_job_statuses
from positron_common.job_api.stream_logs import start_stdout_stream
from positron_common.job_api.get_job import get_job_status
from positron_common.job_api.start_job import start_job
from positron_common.job_api.terminate_job import terminate_job
from positron_common.job_api.create_job import create_job
from positron_common.cli_args import args as cli_args
from positron_common.compression import create_workspace_tar, check_workspace_dir
from positron_common.env_config import env
from positron_common.user_config import user_config
from positron_common.constants import temp_path
from ..cli.console import console, ROBBIE_BLUE
from ..cli.logging_config import logger
from ..job_api.terminate_job import terminate_job
from ..print import print_robbie_configuration_banner, print_job_details_banner, print_known_error

# TODO: update env var to just be base url
PORTAL_BASE = env.API_BASE.rstrip('/api')

class Deploy:
    @staticmethod
    def remote_function_deploy(func, args, kwargs, job_config: PositronJob):
        signal.signal(signal.SIGINT, handle_sigint)
        try:
            # TODO: lots of this code is the same as the other `deploy` file, need to consolidate
            logger.debug(f'Job Config: {job_config}')
            job_config.validate_values()
            logger.debug(f'Runtime Environment: {job_config.create_runtime_env()}')

            job = create_job(job_config)
            print_robbie_configuration_banner(job, job_config)
            logger.debug(f"Created Run Details: {json.dumps(job, indent=2)}")

            # prompt the user if they don't pass the -y option
            if not cli_args.skip_prompts:
                user_input = input("Run with these settings? (Y/n)")
                if not user_input.lower() in ["", "yes", "y", "Yes", "Y"]:
                    terminate_job(job["id"], "User declined from CLI")
                    console.print("[yellow]See you soon![/yellow]")
                    return

            console.print("[bold]Local Machine: [/bold]", style=ROBBIE_BLUE)    
            workspace_dir = (job_config.workspace_dir if job_config.workspace_dir else get_default_workspace_dir())
            logger.debug(f"Workspace Directory: {workspace_dir}")

            if not check_workspace_dir(workspace_dir):
                console.print("[yellow]See you soon![/yellow]")
                terminate_job(job["id"], "User declined run from CLI")
                _exit_by_mode(0)
                return

            # Create stored function from func and arguments
            stored_function = StoredFunction(func, args, kwargs)
            stored_function.set_job_id(job['id'])
            stored_function.serialize_function()
            stored_function.create_function_metadata(hmac_key=user_config.user_auth_token)
            file_count = create_workspace_tar(workspace_dir=workspace_dir)
            console.print("[green]✔[/green] Packaged up workspace artifacts...(1 of 3)", style=ROBBIE_BLUE)

            if file_count == 0:
                Confirm.ask("No files were found in the workspace directory. Would you like to continue anyway?", default=False)

            stored_function.upload_to_s3()
            S3PresignedHandler.upload_file_to_job_folder(f"{temp_path}/{env.COMPRESSED_WS_NAME}", job['id'], env.COMPRESSED_WS_NAME)
            console.print("[green]✔[/green] Uploaded compressed workspace to Robbie...(2 of 3)", style=ROBBIE_BLUE)

            if cli_args.create_only:
                console.print(f"[green]✔[/green] Run created successfully. (3 of 3)")
                console.print(f"JOB_ID: {job.get('id')}, NAME: {job.get('name')}")
                _exit_by_mode(0)
                return

            # start job
            runtime_env = job_config.create_runtime_env()
            logger.debug(f"Runtime Environment: {runtime_env}")
            start_job(job_id=job['id'], data=runtime_env)
            console.print("[green]✔[/green] Submitted run to Robbie...(3 of 3)", style=ROBBIE_BLUE)
            print_job_details_banner(job)
            console.print(f"You can also monitor run status in the Robbie portal at: {PORTAL_BASE}/portal/app/my-runs?jobId={job['id']}\n", style=ROBBIE_BLUE)

            # Start standard output stream if option selected
            console.print("Waiting for remote job to finish...", style=ROBBIE_BLUE)
            if cli_args.stream_stdout:
                start_stdout_stream(job['id'])

                #  Wait for the remote function run to finish
                while True:
                    job = get_job_status(job['id'])
                    if job['status'] == JobStatus.complete:
                        break
                    elif job['status'] in failed_job_statuses:
                        raise RemoteCallException(user_friendly_message="The remote job has failed, please check job output logs for further details!")
                    else:
                        time.sleep(10)
                
                # Job done, handle results
                console.print(f"[green]✔[/green] Done! Now processing results...")
                result, exception = stored_function.load_and_validate_results(hmac_key=user_config.user_auth_token)
                console.print("Passing back result.")
                if exception is not None:
                    raise RemoteFunctionException(exception)
                else:
                    return result
            
        except RemoteCallException as e:
            """For known errors we dont print exceptions, we just print the user friendly message"""
            print_known_error(e)
            _exit_by_mode(1)
        except RemoteFunctionException as e:
            """
                When the result of a remote function call is an Exception being raised
                this is how we let the deserialized exception bubble up back to the client side
            """
            raise e.rf_exception
        except Exception:
            console.print_exception()


def handle_sigint(signum, frame):
    console.print('Terminating gracefully...')
    # TODO: we should actually close open connections
    _exit_by_mode(0)



from setuptools import setup, find_packages

setup(
    name="robbie",
    version="0.1.24",
    description="Run experiments on the Robbie Cloud",
    author="Positron Networks",
    author_email="support@robbie.run",
    url='https://www.robbie.run',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(where="app"),
    package_dir={"": "app"},
    package_data={
        'positron_common': ['build_env'],
    },
    install_requires=[
        'requests>=2.20.0,<3.0.0',
        'python-socketio>=5.0.0,<6.0.0',
        'colorama>=0.4.0,<1.0.0',
        'configparser>=5.0.0,<7.0.0',
        'argparse>=1.4.0',
        'aiohttp>=3.7.4.post0',
        'typer>=0.12.3',
        'pyyaml>=6.0.1',
        'pydantic>=2.8.0',
        'pydantic_core>=2.20.0',
        's3transfer>=0.10.0',
        'boto3>=1.34.0',
        'cloudpickle>=3.0.0',
        'python-dotenv>=1.0.1',
        'rich >= 13.8.1',
        'pyfiglet >= 1.0.2',
        'prompt_toolkit >= 3.0.47',
        'pydantic_yaml >= 1.3.0',
        'nest_asyncio >= 1.6.0',
        'ipywidgets >= 7, <8',
        'pipreqs >= 0.5.0',
        'tqdm >= 4.66.5',
        'dill >= 0.3.4',
        'stdlib_list>=0.11.0',
        'matplotlib>=3.9.2',
        'sentry-sdk>=2.17.0',
        'jupyter-ui-poll>=1.0.0',
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    entry_points={
        'console_scripts': [
            'robbie=positron_cli.cli:app',
            'positron-job-runner=positron_job_runner.cli:app',
        ],
    },
    python_requires='>=3.6',
    license='Apache License 2.0',
)



import os
import traceback
import matplotlib as mpl
from positron_common.deployment.stored_function import StoredFunction
from positron_common.user_config import REMOTE_FUNCTION_SECRET_KEY_NAME
from positron_common.matplotlib_inline.backend_inline import flush_figures

def _get_from_disk(path: str):
  with open(path, "rb") as f:
    return f.read()

def _write_to_disk(path: str, data: bytes):
  with open(path, "wb") as f:
    f.write(data)

def _support_matplotlib():
  """Uses a custom build of backend_inline that will always write plots to disk as PNGs."""
  print('Supporting matplotlib...')
  mpl.use('module://positron_common.matplotlib_inline.backend_inline')

def app():
  """Reads the function parts from disk and executes."""

  _support_matplotlib()

  print('Reading function parts from disk...')
  s_func = _get_from_disk(f"function.pkl")
  s_args = _get_from_disk(f"args.pkl")
  s_kwargs = _get_from_disk(f"kwargs.pkl")

  stored_function = StoredFunction()
  stored_function.s_func = s_func
  stored_function.s_args = s_args
  stored_function.s_kwargs = s_kwargs

  print('Deserializing function and args...')
  stored_function.deserialize_function()

  print('Executing function...')
  stored_function.run()

  print('Flushing figures...')
  flush_figures()

  print('Serializing results...')
  hmac_key = os.environ[REMOTE_FUNCTION_SECRET_KEY_NAME]
  stored_function.serialize_results()
  # Tradeoff, we can put this here to be quicker overall, but it does add procesing
  # on the users dime. Plus, we have to pass down the hmac_key to the process.
  stored_function.create_results_metadata(hmac_key=hmac_key)
  if stored_function.result is not None:
    print(f"Results: {stored_function.result}")
    _write_to_disk(f"result.pkl", stored_function.s_result)
  if stored_function.exception is not None:
    traceback.print_exception(type(stored_function.exception), stored_function.exception, stored_function.exception.__traceback__)
    print(f"Exception: {stored_function.exception}")
    _write_to_disk(f"exception.pkl", stored_function.s_exception)

  results_metadata = stored_function.s_meta.to_json()
  _write_to_disk(f"results_metadata.json", results_metadata)

  print('Done!')

if __name__ == "__main__":
  app()



import requests
import json
from ..exceptions import RemoteCallException
from ..env_config import env
from ..cli.logging_config import logger

# keys
FS_ID="id"
FS_NAME="name"
FS_TOKENS="userTokens"
FS_MENU="menu"
FS_TYPE="type"
FS_DEFAULT_IMAGE_NAME="defaultImageName"
FS_DEFAULT_IMAGE_ID="defaultImageId"
FS_DEF_ENV_NAME="defaultEnvironmentName"
FS_DEF_ENV_ID="defaultEnvironmentId"
FS_PERSONAL_NAME="Personal"
FS_PERSONAL_TYPE="PERSONAL"

# API call
def list_funding_sources():
    Headers = {"PositronAuthToken": env.USER_AUTH_TOKEN  }
    url = f'{env.API_BASE}/list-funding-sources'
    logger.debug(f'list_funding_sources Calling: {url}')
    response = requests.get(url, headers=Headers)
    logger.debug(response)
    if response.status_code != 200:
        body = response.json()
        logger.debug(json.dumps(body, indent=2))
        if body.get('userFriendlyErrorMessage'):
            raise RemoteCallException(body.get('userFriendlyErrorMessage'))
        else:
            raise RemoteCallException(f'list_funding_sources failed with http code: {response.status_code} \n {response.text}')
    else:
        logger.debug(json.dumps(response.json(), indent=2))
        return response.json()
    
# keys
ENV_NAME="name"
# note there is a duplicate "environmentName" key in the API response that points to the same thing
ENV_ID="id"
ENV_TPH="tokensPerHour"
ENV_MENU_ITEM="menu"
ENV_CLUSTER_TYPE="clusterType"
ENV_DESCRIPTION="description"
ENV_GPU_NUMBER="gpuNumber"

# API call
def list_environments(fs_id: str):
    Headers = {"PositronAuthToken": env.USER_AUTH_TOKEN, "fundingGroupId": fs_id  }
    url = f'{env.API_BASE}/list-environments'
    logger.debug(f'Calling: {url}')
    response = requests.get(url, headers=Headers)
    logger.debug(response)
    if response.status_code != 200:
        body = response.json()
        logger.debug(json.dumps(body, indent=2))
        if body.get('userFriendlyErrorMessage'):
            raise RemoteCallException(body.get('userFriendlyErrorMessage'))
        else:
            raise RemoteCallException(f'list_environments failed with http code: {response.status_code} \n {response.text}')
    else:
        logger.debug(json.dumps(response.json(), indent=2))
        return response.json()

# keys  
IMAGE_NAME="imageName"
IMAGE_ID="id"
IMAGE_MENU_ITEM="menu"
IMAGE_DELETED="deleted"

# API call
def list_images(fs_id: str, env_id: str):
    Headers = {"PositronAuthToken": env.USER_AUTH_TOKEN, "fundingGroupId": fs_id, "environmentId": env_id  }
    url = f'{env.API_BASE}/list-images'
    logger.debug(f'Calling: {url}')
    response = requests.get(url, headers=Headers)
    logger.debug(response)
    if response.status_code != 200:
        body = response.json()
        logger.debug(json.dumps(body, indent=2))
        if body.get('userFriendlyErrorMessage'):
            raise RemoteCallException(body.get('userFriendlyErrorMessage'))
        else:
            raise RemoteCallException(f'list_images failed with http code: {response.status_code} \n {response.text}')
    else:
        logger.debug(json.dumps(response.json(), indent=2))
        return response.json()
    






import typer
from typing_extensions import Annotated
from positron_common.observability.main import track_command_usage
from positron_common.user_config import user_config
from positron_common.env_defaults import env_config, EnvType
from positron_common.cli.console import console

@track_command_usage("set_env")
def set_env(
  env: Annotated[str, typer.Argument(help='The environment to set.')] = 'dev',
):
    """
    Set's your robbie deploy env. ADVANCED USE ONLY.
    """
    current = env_config[EnvType(env)]
    user_config.backend_api_base_url = current.api_base
    user_config.backend_ws_base_url = current.ws_base
    user_config.write()
    console.print(f'Successfully updated config.')
    console.print(f'[bold]Env[/bold]: {env}')
    console.print(f'[bold]API[/bold]: {current.api_base}')
    console.print(f'[bold]WS [/bold]: {current.ws_base}')



import requests
import json
import pprint
from typing import Optional, List, Dict, Any
from pydantic import BaseModel, Field
from positron_common.utils import undefined
from positron_common.config import PositronJob
from positron_common.cli_args import args as cli_args
from positron_common.env_config import env
from positron_common.exceptions import RemoteCallException, RobbieException
from positron_common.cli.logging_config import logger
from positron_common.env_defaults import current

class CreateJobBody(BaseModel):
    """
    Maps to the request body of the Create Job API
    """
    userDefinedName: Optional[str] = Field(default=undefined)
    fundingGroupId: Optional[str] = Field(default=undefined)
    imageName: Optional[str] = Field(default=undefined)
    environmentId: Optional[str] = Field(default=undefined)
    jobArguments: Optional[List[str]] = Field(default=undefined)
    entryPoint: Optional[str] = Field(default=undefined)
    commands: Optional[str] = Field(default=undefined)
    maxTokens: Optional[int] = Field(default=undefined)
    maxMinutes: Optional[int] = Field(default=undefined)

    def http_dict(self) -> Dict[str, Any]:
        """
        Enables dropping fields that were never set and should be treated as undefined
        """
        return {k: v for k, v in self.__dict__.items() if v is not undefined}

    @staticmethod
    def from_config(job_config: PositronJob):
        instance = CreateJobBody(
            maxTokens=job_config.max_tokens,
            maxMinutes=job_config.max_time,
        )
        if not job_config.job_type:
            raise RobbieException("Internal error: Job type is not set")
        instance.entryPoint = job_config.job_type.value
        if job_config.name:
            instance.userDefinedName = job_config.name
        if job_config.commands:
            instance.commands = ";\n".join(job_config.commands)
        else:
            instance.jobArguments = cli_args.job_args
        if job_config.funding_group_id:
            instance.fundingGroupId = job_config.funding_group_id
        if job_config.image:
            instance.imageName = job_config.image
        if job_config.environment_id:
            instance.environmentId = job_config.environment_id
        return instance



def create_job(job_config: PositronJob):
    data = CreateJobBody.from_config(job_config)
    logger.debug(pprint.pformat(data.http_dict()))
    url = f'{env.API_BASE}/create-job'
    logger.debug(f'Calling: {url}')
    Headers = {"PositronAuthToken": env.USER_AUTH_TOKEN}
    response = requests.post(url, headers=Headers, json=data.http_dict())
    logger.debug(response)
    if response.status_code != 200:
        body = response.json()
        logger.debug(json.dumps(body, indent=2))
        # TODO: need standard error codes in errors
        if (body.get('message').lower() == 'concurrent job limit exceeded'):
            additional_help = [
                "You can free up capacity by terminating running jobs or reaching out to your group's",
                "administrator to increase your concurrent job limit.",
                f"{current.portal_base}/portal/app/my-runs"
            ]
            raise RemoteCallException(
                reason=body.get('message').lower(),
                user_friendly_message=body.get('userFriendlyErrorMessage'),
                additional_help="\n".join(additional_help)
            )
        if body.get('userFriendlyErrorMessage'):
            raise RemoteCallException(body.get('userFriendlyErrorMessage'))
        else:
            raise RemoteCallException(f'Job creation failed with http code: {response.status_code} \n {response.text}')
    else:
        # logger.debug(response.json())
        logger.debug(json.dumps(response.json(), indent=2))
        return response.json()




from positron_common.utils import get_version
from positron_common.build_env import build_env
from positron_common.env_defaults import current
from positron_common.cli.console import console
from positron_common.observability.main import track_command_usage

@track_command_usage("version")
def version():
    console.print(f"Robbie SDK Version: {get_version()}, build_env: {build_env.value}, current_env: {current.name}")




import json
import sys
import time
import logging
from logging import ERROR, INFO, DEBUG, WARN
from positron_job_runner.runner_env import runner_env
from positron_job_runner.cloudwatch_client import create_client
from positron_job_runner.logging_config import get_logger

console_logger = get_logger()

class CloudLogger:
  """A class to log messages to stdout, stderr, and CloudWatch."""
  stdout = []
  stderr = []
  # cloudwatch_client = None

  def __init__(self):
    # Connect to cloudwatch if enabled
    if runner_env.AWS_JOB_LOG_GROUP_NAME:
      self.__cloudwatch_client = create_client()
    else:
      console_logger.info("CloudWatch logging is disabled.")

  def error(self, message: str):
    self.log(message, ERROR)

  def warn(self, message: str):
    self.log(message, WARN)

  def info(self, message: str):
    self.log(message, INFO)
  
  def debug(self, message: str):
    self.log(message, DEBUG)
  
  def exception(self, *args, **kwargs):
    console_logger.exception(*args, **kwargs)
    self.log(str(args[0]), ERROR)

  def log(self, message: str, level = INFO):
    if isinstance(message, bytes):
        message = message.decode('utf-8')
    
    # Store in mem for later retrieval
    if level == ERROR:
        self.stderr.append(message)
    else:
        self.stdout.append(message)

    # Log to console at the provided level.
    console_logger.log(level = level, msg = message)

    # Log to CloudWatch if enabled.
    if runner_env.AWS_JOB_LOG_GROUP_NAME:
      self._log_to_cloudwatch(message, level)

    # This ensures logs in the right order etc, but may slow down the process
    sys.stdout.flush()
    sys.stderr.flush()

  def end_of_logs_signal(self):
    if runner_env.AWS_JOB_LOG_GROUP_NAME:
      self._log_to_cloudwatch("POSITRON_SIGNAL_EOS")

  def _log_to_cloudwatch(self, message: str, level = INFO):
    timestamp = int(round(time.time() * 1000))
    message = {'timestamp': timestamp, 'log_level': logging.getLevelName(level), 'message': message, 'app_name': 'remote_machine'}
    log_events = [{
        'timestamp': timestamp,
        'message': json.dumps(message),
    }]
    log_stream_name = f'positron-job-{runner_env.JOB_ID}'
    log_group_name = runner_env.AWS_JOB_LOG_GROUP_NAME

    try:
        self.__cloudwatch_client.put_log_events(
            logGroupName=log_group_name,
            logStreamName=log_stream_name,
            logEvents=log_events
        )
    except Exception as e:
        print(f'Failed to submit logs: {e}')
  

logger = CloudLogger()


import webbrowser
import typer
from positron_common.user_config import user_config, DEFAULT_USER_CONFIG_PATH   
from positron_common.cli.logging_config import logger
from positron_common.cli.console import console
from positron_common.constants import APP_NAME, APP_HOME_DIR
from positron_common.auth_api.login import get_device_code_payload, get_user_auth_token, wait_for_access_token
from positron_common.env_config import env
from positron_common.exceptions import RemoteCallException
from positron_common.observability.main import track_command_usage

@track_command_usage("login")
def login() -> None:
    """
    Logs you in to your Robbie account and stores API key on your local machine.
    """
    # Get device code
    try:
        console.print('Authenticating with Robbie...')
        device_code_data = get_device_code_payload()

        # Redirect to login
        console.print("1. If a browser window doesn't automatically launch, navigate to: ", device_code_data['verification_uri_complete'])
        console.print('2. Confirm the following code: ', device_code_data['user_code'])
        console.print('3. Enter your username and password!')
        console.print('')
        webbrowser.open(url=device_code_data['verification_uri_complete'], new=2, autoraise=True)

        # Wait for authentication
        access_token = wait_for_access_token(device_code_data)
        logger.debug(f'Access Token: {access_token}')

        # console.print('Requesting User Auth Token')
        user_token_response_data = get_user_auth_token(access_token)

        console.print(f'[green]Writing {APP_NAME} access token to: {DEFAULT_USER_CONFIG_PATH}')
        save_user_token(user_token_response_data['userAuthToken'])
        # ensure that "env" is updated with the new user token if it was imported previsouly
        env.USER_AUTH_TOKEN = user_config.user_auth_token
    except RemoteCallException as e:
        logger.debug(e, exc_info=True)
        console.print(f"[red]{e.user_friendly_message}")
        raise typer.Exit(code=1)
    except Exception as e:
        logger.debug(e, exc_info=True)
        console.print(f"[red]An error occurred: {e}. If the problem continues, reach out to our support team for help.\nEmail: support@robbie.run[/red]")
        raise typer.Exit(code=1)

def save_user_token(user_token):
    user_config.user_auth_token = user_token
    user_config.write()
    



import io
import os
import subprocess
import signal
import sys
import tempfile
import time
import ipywidgets as widgets
from ipywidgets import HBox, Layout
from nbformat import read, NO_CONVERT
from IPython import get_ipython
from IPython.display import display
from rich.jupyter import print
from positron_common.cli.logging_config import logger
from positron_common.cli.console import console, ROBBIE_DEFAULT, ROBBIE_BLUE
from positron_cli.run import run
from positron_common.job_api.funding_envs_images import *
from positron_common.cli.logging_config import set_log_level

RUN_NAME_HELPER_TEXT = '(Optional customized run name)'
COMING_SOON_MENU_ITEM = 'Customize...(coming soon)'

# Each notebook platform is unique in terms of how to derive the notebook name and how to convert its cells to a python file.
# We support three notebook types:
# 1. Visual Studio Code
# 2. Generic Jupyter Notebook
# 3. Google Colab


def _convert_notebook_to_py_based_on_platform():
    ip = get_ipython()
    if '__vsc_ipynb_file__' in ip.user_ns:
        console.print("[green]✔[/green] Visual Studio Code detected!") 
        return _convert_generic_notebook_to_py(ip.user_ns['__vsc_ipynb_file__'])
    elif 'JPY_SESSION_NAME' in os.environ:
        console.print("[green]✔[/green] Generic Jupyter Notebook detected!")
        return _convert_generic_notebook_to_py(os.getenv('JPY_SESSION_NAME'))
    elif 'google.colab' in str(ip):
        console.print("[green]✔[/green] Google Colab detected!")
        path = "/content/notebook.ipynb"
        _convert_colab_notebook_to_py(path)
    else:
        console.print("[red] Unanble to determine current notebook file name")
        return None


def run_notebook(loglevel: str = None):
    """ 
    Run a notebook as a job on Robbie 
    
    This function will display a UI that allows the user to select the hardware to run the notebook on.
    
    """
    if loglevel:
        set_log_level(loglevel)
    #
    # Left Column
    #
    left_column_title = widgets.HTML(
        value="<b>Robbie can run your notebook as a job!</b>",
    )

    # logo
    file = open(os.path.join(os.path.dirname(__file__), 'img', 'dog.png'), "rb")
    image = file.read()

    logo = widgets.Image(
        value=image,
        format='png',
        width=100,
        height=100,
    )

    style = {'description_width': 'initial'}

    #
    # Middle Column
    #
    middle_column_title = widgets.HTML(
        value="<b>Just choose your hardware!</b>",
    )
     # Environment Drop down
    _drop_dict = _load_envs_for_dropdown()
    if _drop_dict is None:
        console.print("[red] Error loading environments")
        return
    _drop_dict.update({COMING_SOON_MENU_ITEM: None})
    
    hardware_dropdown = widgets.Dropdown(
        options = _drop_dict,
        # value=_drop_dict[1],
        description='Hardware Options:',
        style=style,
        layout={'width': 'max-content'}
    )

    _or_ = widgets.HTML(
        value="           - or - ",
    )

    # robbie does it for you
    """
    auto_checkbox = widgets.Checkbox(
        value=False,
        description='Let Robbie choose the best resources for you',
        disabled=False,
        indent=False
    )
    """

    # goes back to simple mode
    take_me_back_button = widgets.Button(
        description='Take me back',
        disabled=False,
        # button_style='info', # 'success', 'info', 'warning', 'danger' or ''
        tooltip='Go back to simple options',
        style = style
    )

    #
    # Right Column
    #
    right_column_title = widgets.HTML(
        value="<b>And get the ball rolling!</b>",
    )

    # optional name
    optional_name = widgets.Text(
        value=RUN_NAME_HELPER_TEXT,
        # description='Enter a name:',
        disabled=False,
        style=style
    )
    
    # "Run" button
    run_button = widgets.Button(
        description='Run',
        disabled=False,
        button_style='success', # 'success', 'info', 'warning', 'danger' or ''
        tooltip='Your notebook will be run as a job on Robbie',
        icon='play',
        style = style
    )

    # make sure the notebooke is saved
    saved_text = widgets.HTML(
        value="\n\nPlease ensure your notebook is saved before hitting to <b>Run</b> button",
    )

    # save help
    file = open(os.path.join(os.path.dirname(__file__), 'img', 'mac_save.png'), "rb")
    image = file.read()

    mac_short_cut_image = widgets.Image(
        value=image,
        format='png',
        width=30,
        height=21,
    )

    mac_shortcut_text = widgets.HTML(
        value=" on a Mac",
    )
    mac_save = HBox([mac_short_cut_image, mac_shortcut_text])
    # dont' stretch the image
    mac_save.layout.align_items = 'center'

    file = open(os.path.join(os.path.dirname(__file__), 'img', 'windows_save.png'), "rb")
    image = file.read()

    windows_shortcut_image = widgets.Image(
        value=image,
        format='png',
        width=49,
        height=18,
    )
    windows_shortcut_text = widgets.HTML(
        value=" on Windows",
    )

    coming_soon_text = widgets.HTML(
        value="Coming soon!",
    )

    # robbie does it for you
    tail_checkbox = widgets.Checkbox(
        value=False,
        description='Tail the remote machine stdout.',
        disabled=False,
        indent=False,
        style=style
    )
    windows_save = HBox([windows_shortcut_image, windows_shortcut_text])
    # dont' stretch the image
    windows_save.layout.align_items = 'center'

    save_hints = widgets.HBox([mac_save, windows_save])

    # layout
    left_column = widgets.VBox([left_column_title, logo])
    middle_column = widgets.VBox([middle_column_title, hardware_dropdown, tail_checkbox],layout=Layout(margin = '0 20px 0 20px'))
    right_column = widgets.VBox([right_column_title, optional_name, run_button, saved_text, save_hints], layout=Layout(margin = '0 20px 0 20px'))
    ui = widgets.HBox([left_column, middle_column, right_column])

    out = widgets.Output()
    # display everything
    display(ui, out)

 
    def on_take_me_back_button_clicked(change):
        """ When we display the full list of option (all disabled) this button gets gets us back to the simple mode"""
        with out:
            # print("on_take_me_back_clicked() - change:", change)
            middle_column.children = [middle_column_title, hardware_dropdown, tail_checkbox]
    
    take_me_back_button.on_click(on_take_me_back_button_clicked)
    

    # this is the hardware dropdown to change to customization mode
    def on_hw_dropdown_changed(change):
        """ This is experimental now to show users what is coming"""
        with out:
            # print("on_checkbox_clicked() - change:", change)
            if change.get('type') == 'change' and change.get('name') == 'value' and change.get('new') == COMING_SOON_MENU_ITEM:
                gpu_dropdown = widgets.Dropdown(
                    options=['Nvidia A10 (24GB)', 'Nvidia A100 (40GB)', 'Nvidia H100 (40GB)', 'CPU only'],
                    value='Nvidia A100 (40GB)',
                    description='GPU model:',
                    style=style,
                    layout={'width': 'max-content'},
                    disabled=True
                )
                cpu_slider=widgets.IntSlider(description='vCPU cores', min=1, max=64, step=1, value=4, style=style, disabled=True);
                memory_slider=widgets.IntSlider(description='System Memory (GB)', min=16, max=128, step=16, value=32, style=style, disabled=True);
                disksize_slider=widgets.IntSlider(description='Disk Size (GB)', min=16, max=128, step=16, value=32, style=style, disabled=True);
                middle_column.children = [middle_column_title, coming_soon_text, gpu_dropdown, cpu_slider, memory_slider, disksize_slider, take_me_back_button]

    hardware_dropdown.observe(on_hw_dropdown_changed)

    """
        # this is the robbie does it for you checkbox
        def on_checkbox_clicked(change):
            with out:
                # print("on_checkbox_clicked() - change:", change)
                if change.get('type') == 'change' and change.get('name') == 'value' and change.get('new') == True:
                    # print("Checkbox is checked")
                    # gpu_dropdown.disabled = True 
                    # cpu_slider.disabled = True
                    # memory_slider.disabled = True
                    hardware_dropdown.disabled = True
                    # disksize_slider.disabled = True
                elif change.get('type') == 'change' and change.get('name') == 'value' and change.get('new') == False:
                    # print("Checkbox is unchecked")
                    # gpu_dropdown.disabled = False
                    # cpu_slider.disabled = False
                    # memory_slider.disabled = False
                    hardware_dropdown.disabled = False
                    # disksize_slider.disabled = False

        # register the event handler
        auto_checkbox.observe(on_checkbox_clicked)
    """
    def on_button_clicked(b):
        """ The user clicked the run button """
        with out:
            logger.debug(f"Hardware dropdown value: {hardware_dropdown.value}")  
            if hardware_dropdown.value == None:
                console.print("[red] Please choose a hardware option, before running.")
                return
            
            py_path = _convert_notebook_to_py_based_on_platform()
            if py_path:
                console.print("[green]✔[/green] Converting notebook to Python..Success!")
            else:
                console.print("[red]Comverting notebook to Python..failed!")
                return
            
            console.print('Sending run to Robbie...')

            # this was the old way of doing things
            # kept here for reference
            '''
            run_cmd = f'robbie run "pip install uv && uv pip install -r ./requirements.txt && python {py_path}" --environment_id {hardware_dropdown.value} --y --auto-dep'
            if (optional_name.value and optional_name.value != RUN_NAME_HELPER_TEXT):
                run_cmd += f' --name "{optional_name.value}"'
            if tail_checkbox.value:
                run_cmd += ' --tail'
  
            # check if they enter a name
            rc = bash(run_cmd, print_stdout=True, print_stderr=True)
            if(rc != 0):
                console.print("[red] Job submission to Robbie..failed! Return code:", rc)
            '''
            _name = optional_name.value if optional_name.value != RUN_NAME_HELPER_TEXT else None

            run(name_arg=_name,
                funding_arg=None,
                environment_arg=hardware_dropdown.value,
                image_arg=None,
                commands_arg=f"pip install uv && uv pip install -r ./requirements.txt && python {py_path}",
                auto_capture_deps=True,
                tail=tail_checkbox.value,
                status=False,
                skip_prompts=True,
                interactive=False,
                download=False,
                create_only=False)
        
    # register the event handler
    run_button.on_click(on_button_clicked)

class VerboseCalledProcessError(subprocess.CalledProcessError):
    def __str__(self):
        if self.returncode and self.returncode < 0:
            try:
                msg = "Command '%s' died with %r." % (
                    self.cmd, signal.Signals(-self.returncode))
            except ValueError:
                msg = "Command '%s' died with unknown signal %d." % (
                    self.cmd, -self.returncode)
        else:
            msg = "Command '%s' returned non-zero exit status %d." % (
                self.cmd, self.returncode)

        return f'{msg}\n' \
               f'Stdout:\n' \
               f'{self.output}\n' \
               f'Stderr:\n' \
               f'{self.stderr}'


def bash(cmd, print_stdout=True, print_stderr=True):
    stdoutfile = tempfile.mktemp()
    stderrfile = tempfile.mktemp()
    with io.open(stdoutfile, 'w') as stdout, io.open(stderrfile, 'w') as stderr, io.open(stdoutfile, 'r') as all_stdout, io.open(stderrfile, 'r') as all_stderr:
        proc = subprocess.Popen(cmd, stderr=stderr, stdout=stdout, shell=True, universal_newlines=True,
                        executable='/bin/bash')

        stdout_text = ''
        stderr_text = ''
        while proc.poll() is None:
            stdout_line = all_stdout.read();
            stderr_line = all_stderr.read();
            stderr_text += stderr_line
            stdout_text += stdout_line
            sys.stdout.write(stdout_line)
            sys.stderr.write(stderr_line)
            time.sleep(0.5)
        
        stdout_line = all_stdout.read();
        stderr_line = all_stderr.read();
        stderr_text += stderr_line
        stdout_text += stdout_line
        sys.stdout.write(stdout_line)
        sys.stderr.write(stderr_line)

        if proc.wait() != 0:
            raise VerboseCalledProcessError(proc.returncode, cmd, stdout_text, stderr_text)
        return proc.returncode

def _convert_generic_notebook_to_py(nb_full_path):
    """
    Convert a generic notebook to a python file
    - works for VScode and Jupyter notebooks
    - remove things that are not compatible with generic python
    """
    if nb_full_path:
        console.print(f"Reading notebook file: {nb_full_path}")
        with open(nb_full_path) as fp:
            notebook = read(fp, NO_CONVERT)
        cells = notebook['cells']
        code_cells = [c for c in cells if c['cell_type'] == 'code']

        write_file = f"{os.path.splitext(os.path.basename(nb_full_path))[0]}.py"
        console.print(f"Writing Python file: {write_file}")
        with open(write_file, 'w') as fp1:
            for no_cell, cell in enumerate(code_cells):
                # print(f"####### Cell {no_cell} #########")
                # each cell is a string
                my_str = cell['source']

                # split the string into lines so we can look at each one
                list_of_lines = my_str.split('\n')
                clean_list = []
                for line in list_of_lines:
                    # discard magics
                    if line.startswith('%'):
                        # print("found %")
                        continue
                    if line.startswith("%%"):
                        # print("found %%")
                        continue
                    # discard robbie imports
                    if line == "import robbie":
                        continue
                    if line == "robbie.init()":
                        continue
                    # wrap shell commands in os.system
                    if line.startswith("!"):
                        # print("found !")
                        line = line.replace(line[0], "", 1)
                        line = f"import os\nos.system('{line}')"
                    clean_list.append(line)
                my_str = '\n'.join(clean_list) 
                fp1.write(my_str)
                fp1.write("\n")
            #print("")
            return write_file
    return None
    
'''
Special way to convert a Google Colab notebook to a python file
- Google Colab notebooks have a different structure
- We need to use the Google Colab API to fetch the cells
'''
def _convert_colab_notebook_to_py(nb_full_path):
    # fetch the cells from the notebook frontend
    from google.colab import _message
    nb = _message.blocking_request('get_ipynb')
    if nb:
        cells = nb["ipynb"]['cells']
        code_cells = [c for c in cells if c['cell_type'] == 'code']

        print("nb_full_path:", nb_full_path)
        write_file = f"{os.path.splitext(os.path.basename(nb_full_path))[0]}.py"
        with open(write_file, 'w') as fp1:
            # go through the cells
            for no_cell, cell in enumerate(code_cells):
                print(f"####### Cell {no_cell} #########")
                print(cell['source'])
                # each cell is a list of strings
                list_of_lines = cell['source']
                # clean up the lines
                clean_list = []
                for line in list_of_lines:
                    # discard magics
                    if line.startswith('%'):
                        print("found %")
                        continue
                    if line.startswith("%%"):
                        print("found %%")
                        continue
                    # discard robbie imports
                    if line == "import robbie":
                        continue
                    if line == "robbie.init()":
                        continue
                    # wrap shell commands in os.system
                    if line.startswith("!"):
                        print("found !")
                        line = line.replace(line[0], "", 1)
                        line = f"import os\nos.system('{line}')"
                    clean_list.append(line)
                my_str = '\n'.join(clean_list) 
                fp1.write(my_str)
                fp1.write("\n")
            return write_file
    return None


#
# Reads the environments from the PERSONAL funding group
#

def _load_envs_for_dropdown():
    """
    Load the environments for the dropdown list in the cell output
    """
    fs = list_funding_sources()
    if len(fs) == 0:
        return None
    
    personal_fs_id = None
    for key, val in fs.items():
        if (val[FS_TYPE] == FS_PERSONAL_TYPE):
            personal_fs_id = val.get(FS_ID)
    
    if personal_fs_id is None:
        return None
    envs = list_environments(personal_fs_id)
    if len(envs) == 0:
        return None
    
    dropdown_list = {}
    for _, val in envs.items():
        dropdown_list.update({val.get(ENV_NAME) + " - " + val.get(ENV_DESCRIPTION) + f' ({val.get(ENV_TPH)} tokens/hr)' : val.get(ENV_ID)})
    return dropdown_list





    




from typing import Dict
from dataclasses import dataclass
from .build_env import build_env, EnvType
from positron_common.cli.logging_config import logger
from positron_common.cli.console import console, print_boxed_messages

@dataclass
class EnvDefaults():
  name: str
  portal_base: str
  api_base: str
  ws_base: str
  auth0_domain: str
  auth0_audience: str
  auth0_client_id: str

  def to_string(self, include_title: bool = False) -> str:
    message = f"""- EnvDefaults: {self.name}
- portal_base: {self.portal_base}
- api_base: {self.api_base}
- ws_base: {self.ws_base}
- auth0_domain: {self.auth0_domain}
- auth0_audience: {self.auth0_audience}
- auth0_client_id: {self.auth0_client_id}"""
    if include_title:
      return f"========== EnvDefaults ==========\n{message}"
    else:
      return message

    

env_config: Dict[EnvType, EnvDefaults] = {
  EnvType.LOCAL: EnvDefaults(
    name='local',
    portal_base='http://localhost:3000',
    api_base='http://localhost:3002/api',
    ws_base='ws://localhost:3002',
    auth0_domain='dev-k1t01pbanrr04itm.us.auth0.com',
    auth0_client_id='Mr3GW8Ub4e2bLaEvZ5o0XK5pGfEhtH3d',
    auth0_audience='https://localhost/positron/api',
  ),
  EnvType.DEV: EnvDefaults(
    name='dev',
    portal_base='https://dev.positronsupercompute.com',
    api_base='https://dev.positronsupercompute.com/api',
    ws_base='wss://dev.positronsupercompute.com',
    auth0_domain='dev-k1t01pbanrr04itm.us.auth0.com',
    auth0_client_id='Mr3GW8Ub4e2bLaEvZ5o0XK5pGfEhtH3d',
    auth0_audience='https://localhost/positron/api',
  ),
  EnvType.ALPHA: EnvDefaults(
    name='alpha',
    portal_base='https://alpha.positronsupercompute.com',
    api_base='https://alpha.positronsupercompute.com/api',
    ws_base='wss://alpha.positronsupercompute.com',
    auth0_domain='dev-k1t01pbanrr04itm.us.auth0.com',
    auth0_client_id='Mr3GW8Ub4e2bLaEvZ5o0XK5pGfEhtH3d',
    auth0_audience='https://localhost/positron/api',
  ),
  EnvType.BETA: EnvDefaults(
    name='beta',
    portal_base='https://robbie.run',
    api_base='https://robbie.run/api',
    ws_base='wss://robbie.run',
    auth0_domain='positron-beta.us.auth0.com',
    auth0_client_id='ZW0vio95rYfbHrN7kE3PoUXwmPloBw7e',
    auth0_audience='https://beta/positron/api',
  ),
}



current = env_config[EnvType.DEV]
# current = env_config[EnvType.LOCAL]

# Set current based on build
if build_env:
  current = env_config[build_env]
  logger.debug(f'Using build environment: "{build_env.value}" backend url: "{current.api_base}"')
